#ifndef CAFFE_SYNCEDMEM_HPP_
#define CAFFE_SYNCEDMEM_HPP_

#include <cstdlib>
#include <boost/intrusive/list.hpp>

#ifdef USE_MKL
  #include "mkl.h"
#endif

#include "caffe/common.hpp"

namespace caffe {

// If CUDA is available and in GPU mode, host memory will be allocated pinned,
// using cudaMallocHost. It avoids dynamic pinning for transfers (DMA).
// The improvement in performance seems negligible in the single GPU case,
// but might be more significant for parallel training. Most importantly,
// it improved stability for large models on many GPUs.
inline void CaffeMallocHost(void** ptr, size_t size, bool* use_cuda,
                            int* alloc_device) {
#ifndef CPU_ONLY
  if (Caffe::mode() == Caffe::GPU) {
    CUDA_CHECK(cudaGetDevice(alloc_device));
    CUDA_CHECK(cudaMallocHost(ptr, size));
    *use_cuda = true;
    return;
  }
#endif
#ifdef USE_MKL
  *ptr = mkl_malloc(size ? size:1, 64);
#else
  *ptr = malloc(size);
#endif
  *use_cuda = false;
  CHECK(*ptr) << "host allocation of size " << size << " failed";
}

inline void CaffeFreeHost(void* ptr, bool use_cuda, int alloc_device) {
#ifndef CPU_ONLY
  if (use_cuda) {
    int initial_device;
    cudaGetDevice(&initial_device);
    if (alloc_device != -1)
        CUDA_CHECK(cudaSetDevice(alloc_device));
    CUDA_CHECK(cudaFreeHost(ptr));
    cudaSetDevice(initial_device);
    return;
  }
#endif
#ifdef USE_MKL
  mkl_free(ptr);
#else
  free(ptr);
#endif
}

#ifndef CPU_ONLY
class lms_manager;
typedef boost::intrusive::list_member_hook<boost::intrusive::link_mode<boost::intrusive::auto_unlink> > ListHook;
#endif
/**
 * @brief Manages memory allocation and synchronization between the host (CPU)
 *        and device (GPU).
 *
 * TODO(dox): more thorough description.
 *
 * Regular (non-LMS) SyncedMemory objects treat GPU memory as a mirror of CPU memory.  This mirror is reserved
 * upon the initial transfer/instantiation of data on the GPU and not released until object destruction (or gpu_ptr_
 * replacement).
 * [This minimizes CPU <-> GPU transfer time at the expense of GPU memory use.]
 *
 * LMS-enabled SyncedMemory objects treat GPU memory as a pool.  Minimally, this pool contains the current
 * working set of data required at any stage of execution.  The pool's size and contents are managed dynamically
 * and are dictated by the architecture of the network and GPU memory availability.
 * [This allows for GPU memory overcommitment at the expense of additional CPU <-> GPU transfer time.]
 */

class SyncedMemory {
 public:
  SyncedMemory();
  explicit SyncedMemory(size_t size);
  ~SyncedMemory();
  const void* cpu_data();
  void set_cpu_data(void* data);
  const void* gpu_data();
  void set_gpu_data(void* data);
  void* mutable_cpu_data();
  void* mutable_gpu_data();
  enum SyncedHead { UNINITIALIZED, HEAD_AT_CPU, HEAD_AT_GPU, SYNCED };
  SyncedHead head() const { return head_; }
  size_t size() const { return size_; }

#ifndef CPU_ONLY
  void async_gpu_push(const cudaStream_t& stream);

  // LMS API
  enum Direction { Forward, Backward };
  void allow_reclaim(Direction direction);
  void discard_data(Direction direction);
#endif

 private:
  // LMS state
  bool lms_mode_, lms_reclaimed_forward_, lms_reclaimed_backward_, lms_discard_;
#ifndef CPU_ONLY
  enum ReclaimMode { Ready = 0, Eager = 1, Lazy = 2, ReclaimModeCount = 3 };
  lms_manager *lms_mgr_;
  Direction lms_reclaim_dir_;
  ReclaimMode lms_reclaim_mode_;
  ListHook lms_reclaim_list_hook_;
  cudaStream_t lms_stream_;

  // Memory access and disposal
  void* get_gpu_ptr();
  void free_gpu_ptr();
#endif
  void* get_cpu_ptr();
  void free_cpu_ptr();

#ifndef CPU_ONLY
  // LMS reclaim list management
  bool synchronize_reclaim();
  void prevent_reclaim();
  void reclaim();
  bool on_reclaim_list() { return lms_reclaim_list_hook_.is_linked(); }
  bool& reclaim_history() {
    return (lms_reclaim_dir_ == Forward) ? lms_reclaimed_forward_ : lms_reclaimed_backward_;
  }
#endif

  // Debug/test sanity checks
  void check_device();

  enum AccessMode { Read, Write };
  enum CopyMode { Synchronous, Asynchronous };
  void to_cpu(AccessMode access, CopyMode copy = Synchronous);
  void to_gpu(AccessMode access);
  void* cpu_ptr_;
  void* gpu_ptr_;
  size_t size_;
  SyncedHead head_;
  bool own_cpu_data_;
  bool cpu_malloc_use_cuda_;
  bool own_gpu_data_;
  int device_;

  // device used when cpu_ptr_ is allocated
  int  alloc_device_;
  DISABLE_COPY_AND_ASSIGN(SyncedMemory);

  friend class lms_manager; // For reclaim list access
};  // class SyncedMemory

}  // namespace caffe

#endif  // CAFFE_SYNCEDMEM_HPP_
